import math
from typing import List, Tuple, Dict, Union, Set

import numpy as np
from scipy.sparse import csr_matrix
from sklearn.cluster import AgglomerativeClustering
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import FunctionTransformer

from gtcacs.text_compression import GenerativeTextCompressionNN


class GTCACS:
    """
        Generative Text Compression with Agglomerative Clustering Summarization (GTCACS):
        a NLP model for topic extraction, presented in the paper
        'Therapy Analytics using a Patient-centered Perspective: an application to Hypothyroidism'
    """

    def __init__(self,
                 num_topics: int,
                 max_num_words: int = 100,
                 max_df: Union[float, int] = 1.0,
                 min_df: Union[int, float] = 2,
                 stopwords: Set[str] = None,
                 ngram_range: Tuple[int, int] = (1, 1),
                 lowercase: bool = True,
                 max_features: Union[int, None] = None,
                 num_epoches: int = 5,
                 batch_size: int = 64,
                 gen_learning_rate: float = 0.005,
                 discr_learning_rate: float = 0.005,
                 random_seed_size: int = 100,
                 generator_hidden_dim: int = 512,
                 document_dim: int = None,
                 latent_space_dim: int = 64,
                 discriminator_hidden_dim: int = 256):
        """
            Initialization

            Parameters
            ----------
            :param num_topics: number of topics
            :param max_num_words: maximum number of terms to consider
            :param max_df: maximum document frequency
            :param min_df: minimum document frequency
            :param stopwords: stopwords set
            :param ngram_range: range for ngram
            :param lowercase: flag for convert to lowercase
            :param max_features: maximum number of terms to consider (max vocabulary size)
            :param num_epoches: number of epochs
            :param batch_size: number of documents in a batch
            :param gen_learning_rate: learning rate for optimize the generative part
            :param discr_learning_rate: learning rate for optimize the discriminative part
            :param random_seed_size: dimension of generator input layer
            :param generator_hidden_dim: dimension of generator hidden layer
            :param document_dim: dimension of generator output layer and discriminator's input/output layer
            :param latent_space_dim: dimension of discriminator latent space
            :param discriminator_hidden_dim: dimension of discriminator hidden layer
        """

        self.num_topics = num_topics
        self.max_num_words = max_num_words

        # vectorization
        self.max_df = max_df
        self.min_df = min_df
        self.stopwords = stopwords
        self.ngram_range = ngram_range
        self.max_features = max_features
        self.lowercase = lowercase

        # compression NN
        self.num_epoches = num_epoches
        self.batch_size = batch_size
        self.gen_learning_rate = gen_learning_rate
        self.discr_learning_rate = discr_learning_rate
        self.random_seed_size = random_seed_size
        self.generator_hidden_dim = generator_hidden_dim
        self.document_dim = document_dim
        self.latent_space_dim = latent_space_dim
        self.discriminator_hidden_dim = discriminator_hidden_dim

        # states
        self.corpus_transformed_shape = None
        self.is_fitted = False
        self.clusters_labels = None
        self.terms_frequency_map = None
        self.topics_matrix = None
        self.topics_distribution = None

        # models
        self.vectorizer_model = Pipeline(
            steps=[
                ("vect", TfidfVectorizer(ngram_range=self.ngram_range, stop_words=self.stopwords,
                                         max_df=self.max_df, min_df=self.min_df, lowercase=self.lowercase,
                                         max_features=self.max_features)),
                ("dense", FunctionTransformer(csr_matrix.todense))
            ]
        )
        self.dim_red_model = GenerativeTextCompressionNN(
            num_epoches=self.num_epoches,
            batch_size=self.batch_size,
            gen_learning_rate=self.gen_learning_rate,
            discr_learning_rate=self.discr_learning_rate,
            gen_input_random_noise_size=self.random_seed_size,
            gen_hidden1_size=self.generator_hidden_dim,
            gen_output_size=self.document_dim,
            discr_encoder_input_size=self.document_dim,
            discr_encoder_hidden1_size=self.discriminator_hidden_dim,
            discr_encoder_output_size=self.latent_space_dim,
            discr_decoder_input_size=self.latent_space_dim,
            discr_decoder_hidden1_size=self.discriminator_hidden_dim,
            discr_decoder_output_size=self.document_dim
        )
        self.clustering_model = AgglomerativeClustering(
            n_clusters=self.num_topics,
            affinity="euclidean",
            linkage="ward"
        )

    def _build_compression_network(self, num_features: int):
        self.corpus_transformed_shape = num_features
        if (not self.document_dim) or (self.document_dim > self.corpus_transformed_shape):
            self.dim_red_model.build_network(num_features=self.corpus_transformed_shape)
        else:
            self.dim_red_model.build_network(num_features=self.document_dim)

    def extract_topics(self, corpus: List[str]):
        """
            Extract topics from the corpus of documents given in input
            (firstly call this method for fit on corpus)

            :param corpus: list of textual documents
        """

        corpus_transformed = self.vectorizer_model.fit_transform(X=corpus, y=None)
        self._build_compression_network(num_features=corpus_transformed.shape[1])
        self.dim_red_model.train(dataset=corpus_transformed)
        corpus_transformed_compressed = self.dim_red_model.get_latent_space(x_new=corpus_transformed)
        self.clusters_labels = self.clustering_model.fit_predict(X=corpus_transformed_compressed, y=None)
        self.terms_frequency_map = self._compute_terms_frequencies_map(corpus=corpus)
        clusters_partition = self._compute_clusters_partition(corpus=corpus)
        self.topics_matrix = self._compute_topics_matrix(clusters_partition=clusters_partition,
                                                         terms_frequencies_map=self.terms_frequency_map,
                                                         num_top_words=self.max_num_words)
        self.topics_distribution = self._compute_topics_distribution(corpus=corpus)
        self.is_fitted = True

    def _check_is_fitted(self):
        if not self.is_fitted:
            raise ValueError("The topics are not already extracted: call 'extract_topics' first!")

    def get_topics_distribution_scores(self) -> np.ndarray:
        """
            Return the topics distribution scores in the corpus
        """
        self._check_is_fitted()
        return self.topics_distribution

    def get_topics_words(self) -> List[List[Tuple[str, float]]]:
        """
            Return the clusters of terms representing discussion topics
        """
        self._check_is_fitted()
        return self.topics_matrix

    def _compute_topics_distribution(self, corpus: List[str]) -> np.ndarray:
        vec = CountVectorizer(ngram_range=self.ngram_range, stop_words=self.stopwords, lowercase=self.lowercase,
                              max_df=self.max_df, min_df=self.min_df, max_features=self.max_features)
        word_tokenizer_fun = vec.build_tokenizer()
        topics_matrix_tmp = [dict(topic_list) for topic_list in self.topics_matrix]
        result = np.zeros(shape=(len(corpus), self.num_topics))
        for i, doc in enumerate(corpus):
            for token in word_tokenizer_fun(doc):
                for j, topic in enumerate(topics_matrix_tmp):
                    try:
                        result[i, j] += topic[token]
                    except KeyError:
                        continue
        return result

    def _compute_clusters_partition(self, corpus: List[str]) -> Dict[str, List[str]]:
        clusters_partition = dict()
        for i, label in enumerate(self.clusters_labels):
            if label in clusters_partition:
                clusters_partition[label].append(corpus[i])
            else:
                clusters_partition[label] = [corpus[i]]
        return clusters_partition

    def _compute_topics_matrix(self,
                               clusters_partition: Dict[str, List[str]],
                               terms_frequencies_map: Dict[str, int],
                               num_top_words: int = 100) -> List[List[Tuple[str, float]]]:
        topics_matrix = []
        for label, cluster_corpus in clusters_partition.items():
            top_cluster_tokens = self._compute_top_tokens(corpus=cluster_corpus,
                                                          terms_frequencies_map=terms_frequencies_map,
                                                          num_top_words=num_top_words)
            topics_matrix.append([(token, score) for token, score in top_cluster_tokens])
        return topics_matrix

    def _compute_terms_frequencies_map(self, corpus: List[str]) -> Dict[str, int]:
        vec = CountVectorizer(ngram_range=self.ngram_range, stop_words=self.stopwords,
                              lowercase=self.lowercase, max_df=1.0, min_df=1, max_features=None, )
        vec.fit(corpus)
        bag_of_words = vec.transform(corpus)
        sum_words = bag_of_words.sum(axis=0)
        words_freq_map = {word: sum_words[0, idx] for word, idx in vec.vocabulary_.items()}
        return words_freq_map

    def _compute_top_tokens(self,
                            corpus: List[str],
                            terms_frequencies_map: Dict[str, int],
                            num_top_words: int) -> List[Tuple[str, float]]:
        vec = CountVectorizer(ngram_range=self.ngram_range, stop_words=self.stopwords,
                              lowercase=self.lowercase, max_df=1.0, min_df=1, max_features=None)
        vec.fit(corpus)
        bag_of_words = vec.transform(corpus)
        sum_words = bag_of_words.sum(axis=0)
        corpus_words_freq = [(word, sum_words[0, idx]) for word, idx in vec.vocabulary_.items()]
        corpus_words_freq_normalized = [(term, freq / math.sqrt(terms_frequencies_map[term])) for term, freq in
                                        corpus_words_freq]
        corpus_words_freq_normalized_sorted = sorted(corpus_words_freq_normalized, key=lambda x: x[1], reverse=True)
        return corpus_words_freq_normalized_sorted[0:num_top_words]



import logging

import numpy as np
import tensorflow as tf
from tensorflow.python.keras.constraints import MinMaxNorm
from tqdm import tqdm

logging.basicConfig(level=logging.INFO)
tf.keras.backend.set_floatx("float64")
tf.executing_eagerly()


class Generator(tf.keras.Model):

    def __init__(self,
                 gen_input_random_noise_size: int,
                 gen_hidden1_size: int,
                 gen_output_size: int):
        """
            Definition of the Generator Layers
        """
        super().__init__(name='generator')
        self.input_layer = tf.keras.layers.Dense(units=gen_input_random_noise_size,
                                                 activation=tf.nn.leaky_relu,
                                                 kernel_regularizer="l1",
                                                 activity_regularizer="l1")
        self.input_dropout = tf.keras.layers.Dropout(rate=0.1)
        self.hidden1 = tf.keras.layers.Dense(units=gen_hidden1_size,
                                             activation=tf.nn.leaky_relu,
                                             kernel_regularizer="l1",
                                             activity_regularizer="l1")
        self.hidden1_dropout = tf.keras.layers.Dropout(rate=0.2)
        self.output_layer = tf.keras.layers.Dense(units=gen_output_size,
                                                  activation=tf.nn.leaky_relu,
                                                  kernel_regularizer="l1",
                                                  activity_regularizer="l1",
                                                  kernel_constraint=MinMaxNorm(min_value=0.0, max_value=1.0))
        self.output_dropout = tf.keras.layers.Dropout(rate=0.4)

    def call(self, input_tensor, **kwargs):
        """
            Definition of Forward Pass
        """
        x = self.input_layer(input_tensor)
        x = self.input_dropout(x)
        x = self.hidden1(x)
        x = self.hidden1_dropout(x)
        x = self.output_layer(x)
        x = self.output_dropout(x)
        return x

    def generate_noise(self, batch_size: int, random_noise_size: int):
        """
            Method for generate the startup noise input tensor of the generator
        """
        return np.random.uniform(0, 1, size=(batch_size, random_noise_size))


class Discriminator(tf.keras.Model):

    def __init__(self,
                 discr_encoder_input_size: int,
                 discr_encoder_hidden1_size: int,
                 discr_encoder_output_size: int,
                 discr_decoder_input_size: int,
                 discr_decoder_hidden1_size: int,
                 discr_decoder_output_size: int):
        """
            Definition of the Discriminator Layers
        """
        super(Discriminator, self).__init__()

        ### ENCODER ###
        self.encoder_input = tf.keras.layers.InputLayer(input_shape=(discr_encoder_input_size,))  # not indispensable
        # self.encoder_noise = tf.keras.layers.GaussianNoise(stddev=discr_noise_std)
        self.encoder_input_dropout = tf.keras.layers.Dropout(rate=0.4)
        self.encoder_hidden1 = tf.keras.layers.Dense(units=discr_encoder_hidden1_size,
                                                     activation=tf.nn.leaky_relu,
                                                     kernel_regularizer="l1",
                                                     activity_regularizer="l1")
        self.encoder_hidden1_dropout = tf.keras.layers.Dropout(rate=0.1)
        self.encoder_output = tf.keras.layers.Dense(units=discr_encoder_output_size,
                                                    activation=tf.nn.leaky_relu,
                                                    kernel_regularizer="l1",
                                                    activity_regularizer="l1",
                                                    # kernel_constraint=MinMaxNorm(min_value=-1.0, max_value=1.0)
                                                    )

        ### DECODER ###
        self.decoder_input = tf.keras.layers.Input(shape=(discr_decoder_input_size,))  # not indispensable
        self.decoder_hidden1 = tf.keras.layers.Dense(units=discr_decoder_hidden1_size,
                                                     activation=tf.nn.leaky_relu,
                                                     kernel_regularizer="l1",
                                                     activity_regularizer="l1")
        self.decoder_hidden1_dropout = tf.keras.layers.Dropout(rate=0.1)
        self.decoder_output = tf.keras.layers.Dense(units=discr_decoder_output_size,
                                                    activation=tf.nn.leaky_relu,
                                                    kernel_regularizer="l1",
                                                    activity_regularizer="l1",
                                                    kernel_constraint=MinMaxNorm(min_value=0.0, max_value=1.0))
        self.decoder_output_dropout = tf.keras.layers.Dropout(rate=0.4)

    def encode(self, inputs):
        x = self.encoder_input_dropout(inputs)
        x = self.encoder_hidden1(x)
        x = self.encoder_hidden1_dropout(x)
        x = self.encoder_output(x)
        return x

    def decode(self, inputs):
        x = self.decoder_hidden1(inputs)
        x = self.decoder_hidden1_dropout(x)
        x = self.decoder_output(x)
        x = self.decoder_output_dropout(x)
        return x

    def call(self, x_train, **kwargs):
        latent_representation = self.encode(x_train)
        output = self.decode(latent_representation)
        return output


class GenerativeTextCompressionNN(tf.keras.Model):
    """
        Generative Text Compression neural network:
        a custom network for dimensional reduction of textual documents based on the concepts of GAN and autoencoder
    """

    def __init__(self,
                 num_epoches: int,
                 batch_size: int,
                 gen_learning_rate: float,
                 discr_learning_rate: float,
                 gen_input_random_noise_size: int,
                 gen_hidden1_size: int,
                 gen_output_size: int,
                 discr_encoder_input_size: int,
                 discr_encoder_hidden1_size: int,
                 discr_encoder_output_size: int,
                 discr_decoder_input_size: int,
                 discr_decoder_hidden1_size: int,
                 discr_decoder_output_size: int):
        super().__init__(name="gan")
        self.num_epoches = num_epoches
        self.batch_size = batch_size
        self.gen_learning_rate = gen_learning_rate
        self.discr_learning_rate = discr_learning_rate
        self.gen_input_random_noise_size = gen_input_random_noise_size
        self.gen_hidden1_size = gen_hidden1_size
        self.gen_output_size = gen_output_size
        self.discr_encoder_input_size = discr_encoder_input_size
        self.discr_encoder_hidden1_size = discr_encoder_hidden1_size
        self.discr_encoder_output_size = discr_encoder_output_size
        self.discr_decoder_input_size = discr_decoder_input_size
        self.discr_decoder_hidden1_size = discr_decoder_hidden1_size
        self.discr_decoder_output_size = discr_decoder_output_size
        self.generator_optimizer = tf.optimizers.Adam(learning_rate=gen_learning_rate, beta_1=0.5)
        self.discriminator_optimizer = tf.optimizers.Adam(learning_rate=discr_learning_rate, beta_1=0.5)
        self.is_trained = False
        self.initial_random_noise = gen_input_random_noise_size
        self.generator = None
        self.discriminator = None
        self.is_built = False

    def _set_item_size(self, num_features: int):
        self.gen_output_size = num_features
        self.discr_encoder_input_size = num_features
        self.discr_decoder_output_size = num_features

    def _check_built_status(self):
        if not self.is_built:
            raise Exception("The network has not yet been built!")

    def build_network(self, num_features: int):
        self._set_item_size(num_features)
        self.generator = Generator(gen_input_random_noise_size=self.gen_input_random_noise_size,
                                   gen_hidden1_size=self.gen_hidden1_size,
                                   gen_output_size=self.gen_output_size)
        print(self.discr_encoder_input_size)
        self.discriminator = Discriminator(discr_encoder_input_size=self.discr_encoder_input_size,
                                           discr_encoder_hidden1_size=self.discr_encoder_hidden1_size,
                                           discr_encoder_output_size=self.discr_encoder_output_size,
                                           discr_decoder_input_size=self.discr_decoder_input_size,
                                           discr_decoder_hidden1_size=self.discr_decoder_hidden1_size,
                                           discr_decoder_output_size=self.discr_decoder_output_size)
        self.is_built = True

    def compute_generator_loss(self, generated_input, fake_output):
        cos_sim = tf.keras.losses.CosineSimilarity()
        fake_loss = cos_sim(generated_input, fake_output)
        return fake_loss

    def compute_discriminator_loss(self, real_input, real_output, generated_input, fake_output):
        cos_sim = tf.keras.losses.CosineSimilarity()
        real_loss = cos_sim(real_input, real_output)
        fake_loss = cos_sim(generated_input, fake_output)
        total_loss = -(tf.abs(real_loss) + tf.abs(fake_loss))
        return total_loss

    def training_step(self,
                      batch: np.ndarray,
                      batch_size: int):
        self._check_built_status()
        with tf.GradientTape() as gen_tape, tf.GradientTape() as disc_tape:
            ### Produce startup noise tensor ###
            noise = self.generator.generate_noise(batch_size=batch_size,
                                                  random_noise_size=self.initial_random_noise)

            ### Generate synthetic sample with the generator ###
            generated_batch = self.generator(noise)  # training=True

            ### Calculate 'Real Output' / 'Fake Output' tensors for the discriminator ###
            real_discr_output = self.discriminator(batch)  # training=True
            # print("\n real:", real_output.numpy().shape)
            fake_discr_output = self.discriminator(generated_batch)  # training=True
            # print("\n fake:", fake_output.numpy().shape)

            ### Compute Cost Functions Losses ###
            generator_loss = self.compute_generator_loss(generated_input=generated_batch,
                                                         fake_output=fake_discr_output)
            discriminator_loss = self.compute_discriminator_loss(real_input=batch,
                                                                 real_output=real_discr_output,
                                                                 generated_input=generated_batch,
                                                                 fake_output=fake_discr_output)

            ### Calculate Gradients ###
            gradients_of_generator = gen_tape.gradient(generator_loss, self.generator.trainable_variables)
            gradients_of_discriminator = disc_tape.gradient(discriminator_loss, self.discriminator.trainable_variables)

            ### Apply Gradients ###
            self.generator_optimizer.apply_gradients(zip(gradients_of_generator,
                                                         self.generator.trainable_variables))
            self.discriminator_optimizer.apply_gradients(zip(gradients_of_discriminator,
                                                             self.discriminator.trainable_variables))
            return generator_loss, discriminator_loss

    def train(self, dataset):
        self._check_built_status()
        logging.info("Start of training process:")
        # Iterate over the epochs
        for i, epoch in enumerate(range(self.num_epoches)):
            logging.info(">>> EPOCH " + str(i + 1))
            # Iterate over the batches
            epoch_generator_losses = []
            epoch_discriminator_losses = []
            for x in tqdm(range(0, len(dataset), self.batch_size)):
                batch = np.array(dataset[x: x + self.batch_size, :])
                generator_loss, discriminator_loss = self.training_step(batch=batch, batch_size=self.batch_size)
                epoch_generator_losses.append(generator_loss.numpy())
                epoch_discriminator_losses.append(discriminator_loss.numpy())
            logging.info("\t\t epoch mean generator loss: " + '{:12f}'.format(np.mean(epoch_generator_losses)) +
                         "  (-/+" + '{:12f}'.format(np.std(epoch_generator_losses)) + ")")
            logging.info("\t\t epoch mean discriminator loss: " + '{:12f}'.format(np.mean(epoch_discriminator_losses)) +
                         "  (-/+" + '{:12f}'.format(np.std(epoch_discriminator_losses)) + ")")
        self.is_trained = True

    def generate_synthetic_samples(self, x_new: np.ndarray) -> np.ndarray:
        self._check_built_status()
        if not self.is_trained:
            raise Exception("The Generative Adversarial Network is not trained!")
        return self.generator(x_new).numpy()

    def get_latent_space(self, x_new) -> np.ndarray:
        """ Getting the compressed Latent Space of the new data given in input """
        self._check_built_status()
        if not self.is_trained:
            raise Exception("The autoencoder is not trained!")
        return self.discriminator.encode(x_new).numpy()



from distutils.core import setup

# read the contents of your README file
from os import path
content_root_directory = path.abspath(path.dirname(__file__))
with open(path.join(content_root_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

VERSION = '0.0.3'

setup(
    name='gtcacs',  # How you named your package folder (MyLib)
    packages=['gtcacs'],  # Chose the same as "name"
    version=VERSION,
    license='MIT',
    description='Generative Text Compression with Agglomerative Clustering Summarization (GTCACS)',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Andrea Lenzi',
    author_email='andrealenzi11@gmail.com',
    url='https://github.com/andrealenzi11/gen-text-compr-aggl-clust-sum.git',
    download_url=f'https://github.com/andrealenzi11/gen-text-compr-aggl-clust-sum/archive/{VERSION}.tar.gz',
    keywords=[
        'discussion topics',
        'topic modeling',
        'topic modelling',
        'topic extraction',
    ],
    install_requires=[
        'tensorflow==2.2.0',
        'scikit-learn==0.23.1',
        'numpy==1.19.1',
        'scipy==1.4.1',
        'tqdm==4.48.0',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',  # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',  # Again, pick a license
        'Programming Language :: Python :: 3',  # Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)
